#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
from unittest import mock

from barbicanclient import exceptions

from heat.common import exception
from heat.engine.clients.os import barbican
from heat.tests import common
from heat.tests import utils


class BarbicanClientPluginTest(common.HeatTestCase):

    def setUp(self):
        super(BarbicanClientPluginTest, self).setUp()
        self.barbican_client = mock.MagicMock()
        con = utils.dummy_context()
        c = con.clients
        self.barbican_plugin = c.client_plugin('barbican')
        self.barbican_plugin.client = lambda: self.barbican_client

    def test_create(self):
        context = utils.dummy_context()
        plugin = context.clients.client_plugin('barbican')
        client = plugin.client()
        self.assertIsNotNone(client.orders)

    def test_get_secret_by_ref(self):
        secret = collections.namedtuple('Secret', ['name'])('foo')
        self.barbican_client.secrets.get.return_value = secret
        self.assertEqual(secret,
                         self.barbican_plugin.get_secret_by_ref("secret"))

    def test_get_secret_payload_by_ref(self):
        payload_content = 'payload content'
        secret = collections.namedtuple(
            'Secret', ['name', 'payload'])('foo', payload_content)
        self.barbican_client.secrets.get.return_value = secret
        expect = payload_content
        self.assertEqual(expect,
                         self.barbican_plugin.get_secret_payload_by_ref(
                             "secret"))

    def test_get_secret_payload_by_ref_not_found(self):
        exc = exceptions.HTTPClientError(message="Not Found", status_code=404)
        self.barbican_client.secrets.get.side_effect = exc
        self.assertRaises(
            exception.EntityNotFound,
            self.barbican_plugin.get_secret_payload_by_ref,
            "secret")

    def test_get_secret_by_ref_not_found(self):
        exc = exceptions.HTTPClientError(message="Not Found", status_code=404)
        self.barbican_client.secrets.get.side_effect = exc
        self.assertRaises(
            exception.EntityNotFound,
            self.barbican_plugin.get_secret_by_ref,
            "secret")


class SecretConstraintTest(common.HeatTestCase):

    def setUp(self):
        super(SecretConstraintTest, self).setUp()
        self.ctx = utils.dummy_context()
        self.mock_get_secret_by_ref = mock.Mock()
        self.ctx.clients.client_plugin(
            'barbican').get_secret_by_ref = self.mock_get_secret_by_ref
        self.constraint = barbican.SecretConstraint()

    def test_validation(self):
        self.mock_get_secret_by_ref.return_value = {}
        self.assertTrue(self.constraint.validate("foo", self.ctx))

    def test_validation_error(self):
        self.mock_get_secret_by_ref.side_effect = exception.EntityNotFound(
            entity='Secret', name='bar')
        self.assertFalse(self.constraint.validate("bar", self.ctx))


class ContainerConstraintTest(common.HeatTestCase):

    def setUp(self):
        super(ContainerConstraintTest, self).setUp()
        self.ctx = utils.dummy_context()
        self.mock_get_container_by_ref = mock.Mock()
        self.ctx.clients.client_plugin(
            'barbican').get_container_by_ref = self.mock_get_container_by_ref
        self.constraint = barbican.ContainerConstraint()

    def test_validation(self):
        self.mock_get_container_by_ref.return_value = {}
        self.assertTrue(self.constraint.validate("foo", self.ctx))

    def test_validation_error(self):
        self.mock_get_container_by_ref.side_effect = exception.EntityNotFound(
            entity='Container', name='bar')
        self.assertFalse(self.constraint.validate("bar", self.ctx))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

cfg.CONF.import_opt('stack_scheduler_hints', 'heat.common.config')


class SchedulerHintsMixin(object):
    """Utility class to encapsulate Scheduler Hint related logic."""

    HEAT_ROOT_STACK_ID = 'heat_root_stack_id'
    HEAT_STACK_ID = 'heat_stack_id'
    HEAT_STACK_NAME = 'heat_stack_name'
    HEAT_PATH_IN_STACK = 'heat_path_in_stack'
    HEAT_RESOURCE_NAME = 'heat_resource_name'
    HEAT_RESOURCE_UUID = 'heat_resource_uuid'

    @staticmethod
    def _path_in_stack(stack):
        # Note: scheduler_hints can only be of DictOfListOfStrings.
        # Convert the list of tuples to list of delimited strings.
        path = []
        for parent_res_name, stack_name in stack.path_in_stack():
            if parent_res_name is not None:
                path.append(','.join([parent_res_name, stack_name]))
            else:
                path.append(stack_name)

        return path

    def _scheduler_hints(self, scheduler_hints):
        """Augment scheduler hints with supplemental content."""
        if cfg.CONF.stack_scheduler_hints:
            if scheduler_hints is None:
                scheduler_hints = {}
            stack = self.stack
            scheduler_hints[self.HEAT_ROOT_STACK_ID] = stack.root_stack_id()
            scheduler_hints[self.HEAT_STACK_ID] = stack.id
            scheduler_hints[self.HEAT_STACK_NAME] = stack.name
            scheduler_hints[
                self.HEAT_PATH_IN_STACK] = self._path_in_stack(stack)
            scheduler_hints[self.HEAT_RESOURCE_NAME] = self.name
            scheduler_hints[self.HEAT_RESOURCE_UUID] = self.uuid
        return scheduler_hints



#
# Copyright (c) 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Based on glance/api/policy.py
"""Policy Engine For Heat."""

from oslo_config import cfg
from oslo_log import log as logging
from oslo_policy import policy
from oslo_utils import excutils

from heat.common import exception
from heat.common.i18n import _
from heat import policies


CONF = cfg.CONF
LOG = logging.getLogger(__name__)

DEFAULT_RULES = policy.Rules.from_dict({'default': '!'})
DEFAULT_RESOURCE_RULES = policy.Rules.from_dict({'default': '@'})

ENFORCER = None


class Enforcer(object):
    """Responsible for loading and enforcing rules."""

    def __init__(self, scope='heat', exc=exception.Forbidden,
                 default_rule=DEFAULT_RULES['default'], policy_file=None):
        self.scope = scope
        self.exc = exc
        self.default_rule = default_rule
        self.enforcer = policy.Enforcer(
            CONF, default_rule=default_rule, policy_file=policy_file)
        self.log_not_registered = True

        # register rules
        self.enforcer.register_defaults(policies.list_rules())
        self.file_rules = self.enforcer.file_rules
        self.registered_rules = self.enforcer.registered_rules

    def set_rules(self, rules, overwrite=True):
        """Create a new Rules object based on the provided dict of rules."""
        rules_obj = policy.Rules(rules, self.default_rule)
        self.enforcer.set_rules(rules_obj, overwrite)

    def load_rules(self, force_reload=False):
        """Set the rules found in the json file on disk."""
        self.enforcer.load_rules(force_reload)

    def _check(self, context, rule, target, exc,
               is_registered_policy=False, *args, **kwargs):
        """Verifies that the action is valid on the target in this context.

           :param context: Heat request context
           :param rule: String representing the action to be checked
           :param target: Dictionary representing the object of the action.
           :raises heat.common.exception.Forbidden: When permission is denied
                   (or self.exc if supplied).
           :returns: A non-False value if access is allowed.
        """
        do_raise = False if not exc else True
        credentials = context.to_policy_values()
        if is_registered_policy:
            try:
                return self.enforcer.authorize(rule, target, credentials,
                                               do_raise=do_raise,
                                               exc=exc, action=rule)
            except policy.PolicyNotRegistered:
                if self.log_not_registered:
                    with excutils.save_and_reraise_exception():
                        LOG.exception(_('Policy not registered.'))
                else:
                    raise
        else:
            return self.enforcer.enforce(rule, target, credentials,
                                         do_raise, exc=exc, *args, **kwargs)

    def enforce(self, context, action, scope=None, target=None,
                is_registered_policy=False):
        """Verifies that the action is valid on the target in this context.

           :param context: Heat request context
           :param action: String representing the action to be checked
           :param target: Dictionary representing the object of the action.
           :raises heat.common.exception.Forbidden: When permission is denied
                   (or self.exc if supplied).
           :returns: A non-False value if access is allowed.
        """
        _action = '%s:%s' % (scope or self.scope, action)
        _target = target or {}
        return self._check(context, _action, _target, self.exc, action=action,
                           is_registered_policy=is_registered_policy)

    def check_is_admin(self, context):
        """Whether or not is admin according to policy.

        By default the rule will check whether or not roles contains
        'admin' role and is admin project.

           :param context: Heat request context
           :returns: A non-False value if the user is admin according to policy
        """
        return self._check(context, 'context_is_admin', target={}, exc=None,
                           is_registered_policy=True)


def get_policy_enforcer():
    # This method is used by oslopolicy CLI scripts to generate policy
    # files from overrides on disk and defaults in code.
    CONF([], project='heat')
    return get_enforcer()


def get_enforcer():
    global ENFORCER
    if ENFORCER is None:
        ENFORCER = Enforcer()
    return ENFORCER


class ResourceEnforcer(Enforcer):
    def __init__(self, default_rule=DEFAULT_RESOURCE_RULES['default'],
                 **kwargs):
        super(ResourceEnforcer, self).__init__(
            default_rule=default_rule, **kwargs)
        self.log_not_registered = False

    def _enforce(self, context, res_type, scope=None, target=None,
                 is_registered_policy=False):
        try:
            result = super(ResourceEnforcer, self).enforce(
                context, res_type,
                scope=scope or 'resource_types',
                target=target, is_registered_policy=is_registered_policy)
        except policy.PolicyNotRegistered:
            result = True
        except self.exc as ex:
            LOG.info(str(ex))
            raise
        if not result:
            if self.exc:
                raise self.exc(action=res_type)
        return result

    def enforce(self, context, res_type, scope=None, target=None,
                is_registered_policy=False):
        # NOTE(pas-ha): try/except just to log the exception
        result = self._enforce(context, res_type, scope, target,
                               is_registered_policy=is_registered_policy)

        if result:
            # check for wildcard resource types
            subparts = res_type.split("::")[:-1]
            subparts.append('*')
            res_type_wc = "::".join(subparts)
            try:
                return self._enforce(context, res_type_wc, scope, target,
                                     is_registered_policy=is_registered_policy)
            except self.exc:
                raise self.exc(action=res_type)

        return result

    def enforce_stack(self, stack, scope=None, target=None,
                      is_registered_policy=False):
        for res in stack.resources.values():
            self.enforce(stack.context, res.type(), scope=scope, target=target,
                         is_registered_policy=is_registered_policy)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
from unittest import mock

from heat.common import exception as exc
from heat.common import template_format
from heat.engine import stack
from heat.engine import template
from heat.tests import common
from heat.tests import utils


class SoftwareComponentTest(common.HeatTestCase):

    def setUp(self):
        super(SoftwareComponentTest, self).setUp()
        self.ctx = utils.dummy_context()

        tpl = '''
        heat_template_version: 2013-05-23
        resources:
          mysql_component:
            type: OS::Heat::SoftwareComponent
            properties:
              configs:
                - actions: [CREATE]
                  config: |
                    #!/bin/bash
                    echo "Create MySQL"
                  tool: script
                - actions: [UPDATE]
                  config: |
                    #!/bin/bash
                    echo "Update MySQL"
                  tool: script
              inputs:
                - name: mysql_port
              outputs:
                - name: root_password
        '''

        self.template = template_format.parse(tpl)
        self.stack = stack.Stack(
            self.ctx, 'software_component_test_stack',
            template.Template(self.template))
        self.component = self.stack['mysql_component']
        self.rpc_client = mock.MagicMock()
        self.component._rpc_client = self.rpc_client

        @contextlib.contextmanager
        def exc_filter(*args):
            try:
                yield
            except exc.NotFound:
                pass

        self.rpc_client.ignore_error_by_name.side_effect = exc_filter

    def test_handle_create(self):
        config_id = 'c8a19429-7fde-47ea-a42f-40045488226c'
        value = {'id': config_id}
        self.rpc_client.create_software_config.return_value = value
        props = dict(self.component.properties)
        self.component.handle_create()
        self.rpc_client.create_software_config.assert_called_with(
            self.ctx,
            group='component',
            name=None,
            inputs=props['inputs'],
            outputs=props['outputs'],
            config={'configs': props['configs']},
            options=None)
        self.assertEqual(config_id, self.component.resource_id)

    def test_handle_delete(self):
        self.resource_id = None
        self.assertIsNone(self.component.handle_delete())
        config_id = 'c8a19429-7fde-47ea-a42f-40045488226c'
        self.component.resource_id = config_id
        self.rpc_client.delete_software_config.return_value = None
        self.assertIsNone(self.component.handle_delete())
        self.rpc_client.delete_software_config.side_effect = exc.NotFound
        self.assertIsNone(self.component.handle_delete())

    def test_resolve_attribute(self):
        self.assertIsNone(self.component._resolve_attribute('others'))
        self.component.resource_id = None
        self.assertIsNone(self.component._resolve_attribute('configs'))
        self.component.resource_id = 'c8a19429-7fde-47ea-a42f-40045488226c'
        configs = self.template['resources']['mysql_component'
                                             ]['properties']['configs']
        # configs list is stored in 'config' property of SoftwareConfig
        value = {'config': {'configs': configs}}
        self.rpc_client.show_software_config.return_value = value
        self.assertEqual(configs, self.component._resolve_attribute('configs'))
        self.rpc_client.show_software_config.side_effect = exc.NotFound
        self.assertIsNone(self.component._resolve_attribute('configs'))


class SoftwareComponentValidationTest(common.HeatTestCase):

    scenarios = [
        (
            'component_full',
            dict(snippet='''
                 component:
                   type: OS::Heat::SoftwareComponent
                   properties:
                     configs:
                       - actions: [CREATE]
                         config: |
                           #!/bin/bash
                           echo CREATE $foo
                         tool: script
                     inputs:
                       - name: foo
                     outputs:
                       - name: bar
                     options:
                       opt1: blah
                 ''',
                 err=None,
                 err_msg=None)
        ),
        (
            'no_input_output_options',
            dict(snippet='''
                 component:
                   type: OS::Heat::SoftwareComponent
                   properties:
                     configs:
                       - actions: [CREATE]
                         config: |
                           #!/bin/bash
                           echo CREATE $foo
                         tool: script
                 ''',
                 err=None,
                 err_msg=None)
        ),
        (
            'wrong_property_config',
            dict(snippet='''
                 component:
                   type: OS::Heat::SoftwareComponent
                   properties:
                     config: #!/bin/bash
                     configs:
                       - actions: [CREATE]
                         config: |
                           #!/bin/bash
                           echo CREATE $foo
                         tool: script
                 ''',
                 err=exc.StackValidationFailed,
                 err_msg='Unknown Property config')
        ),
        (
            'missing_configs',
            dict(snippet='''
                 component:
                   type: OS::Heat::SoftwareComponent
                   properties:
                     inputs:
                       - name: foo
                 ''',
                 err=exc.StackValidationFailed,
                 err_msg='Property configs not assigned')
        ),
        (
            'empty_configs',
            dict(snippet='''
                 component:
                   type: OS::Heat::SoftwareComponent
                   properties:
                     configs:
                 ''',
                 err=exc.StackValidationFailed,
                 err_msg='resources.component.properties.configs: '
                         'length (0) is out of range (min: 1, max: None)')
        ),
        (
            'invalid_configs',
            dict(snippet='''
                 component:
                   type: OS::Heat::SoftwareComponent
                   properties:
                     configs:
                       actions: [CREATE]
                       config: #!/bin/bash
                       tool: script
                 ''',
                 err=exc.StackValidationFailed,
                 err_msg='is not a list')
        ),
        (
            'config_empty_actions',
            dict(snippet='''
                 component:
                   type: OS::Heat::SoftwareComponent
                   properties:
                     configs:
                       - actions: []
                         config: #!/bin/bash
                         tool: script
                 ''',
                 err=exc.StackValidationFailed,
                 err_msg='component.properties.configs[0].actions: '
                         'length (0) is out of range (min: 1, max: None)')
        ),
        (
            'multiple_configs_per_action_single',
            dict(snippet='''
                 component:
                   type: OS::Heat::SoftwareComponent
                   properties:
                     configs:
                       - actions: [CREATE]
                         config: #!/bin/bash
                         tool: script
                       - actions: [CREATE]
                         config: #!/bin/bash
                         tool: script
                 ''',
                 err=exc.StackValidationFailed,
                 err_msg='Defining more than one configuration for the same '
                         'action in SoftwareComponent "component" is not '
                         'allowed.')
        ),
        (
            'multiple_configs_per_action_overlapping_list',
            dict(snippet='''
                 component:
                   type: OS::Heat::SoftwareComponent
                   properties:
                     configs:
                       - actions: [CREATE, UPDATE, RESUME]
                         config: #!/bin/bash
                         tool: script
                       - actions: [UPDATE]
                         config: #!/bin/bash
                         tool: script
                 ''',
                 err=exc.StackValidationFailed,
                 err_msg='Defining more than one configuration for the same '
                         'action in SoftwareComponent "component" is not '
                         'allowed.')
        ),
    ]

    def setUp(self):
        super(SoftwareComponentValidationTest, self).setUp()
        self.ctx = utils.dummy_context()

        tpl = '''
        heat_template_version: 2013-05-23
        resources:
          %s
        ''' % self.snippet

        self.template = template_format.parse(tpl)
        self.stack = stack.Stack(
            self.ctx, 'software_component_test_stack',
            template.Template(self.template))
        self.component = self.stack['component']
        self.component._rpc_client = mock.MagicMock()

    def test_properties_schema(self):
        if self.err:
            err = self.assertRaises(self.err, self.stack.validate)
            if self.err_msg:
                self.assertIn(self.err_msg, str(err))
        else:
            self.assertIsNone(self.stack.validate())



# Copyright (c) 2016 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import eventlet.queue
import functools

from oslo_log import log as logging
from oslo_utils import excutils

from heat.common import exception
from heat.engine import resource
from heat.engine import scheduler
from heat.engine import stack as parser
from heat.engine import sync_point
from heat.objects import resource as resource_objects
from heat.rpc import api as rpc_api
from heat.rpc import listener_client

LOG = logging.getLogger(__name__)


class CancelOperation(BaseException):
    """Exception to cancel an in-progress operation on a resource.

    This exception is raised when operations on a resource are cancelled.
    """
    def __init__(self):
        return super(CancelOperation, self).__init__('user triggered cancel')


class CheckResource(object):

    def __init__(self,
                 engine_id,
                 rpc_client,
                 thread_group_mgr,
                 msg_queue,
                 input_data):
        self.engine_id = engine_id
        self._rpc_client = rpc_client
        self.thread_group_mgr = thread_group_mgr
        self.msg_queue = msg_queue
        self.input_data = input_data

    def _stale_resource_needs_retry(self, cnxt, rsrc, prev_template_id):
        """Determine whether a resource needs retrying after failure to lock.

        Return True if we need to retry the check operation because of a
        failure to acquire the lock. This can be either because the engine
        holding the lock is no longer working, or because no other engine had
        locked the resource and the data was just out of date.

        In the former case, the lock will be stolen and the resource status
        changed to FAILED.
        """
        fields = {'current_template_id', 'engine_id'}
        rs_obj = resource_objects.Resource.get_obj(cnxt,
                                                   rsrc.id,
                                                   refresh=True,
                                                   fields=fields)
        if rs_obj.engine_id not in (None, self.engine_id):
            if not listener_client.EngineListenerClient(
                    rs_obj.engine_id).is_alive(cnxt):
                # steal the lock.
                rs_obj.update_and_save({'engine_id': None})

                # set the resource state as failed
                status_reason = ('Worker went down '
                                 'during resource %s' % rsrc.action)
                rsrc.state_set(rsrc.action,
                               rsrc.FAILED,
                               str(status_reason))
                return True
        elif (rs_obj.engine_id is None and
              rs_obj.current_template_id == prev_template_id):
            LOG.debug('Resource id=%d stale; retrying check', rsrc.id)
            return True
        LOG.debug('Resource id=%d modified by another traversal', rsrc.id)
        return False

    def _handle_resource_failure(self, cnxt, is_update, rsrc_id,
                                 stack, failure_reason):
        failure_handled = stack.mark_failed(failure_reason)
        if not failure_handled:
            # Another concurrent update has taken over. But there is a
            # possibility for that update to be waiting for this rsrc to
            # complete, hence retrigger current rsrc for latest traversal.
            self._retrigger_new_traversal(cnxt, stack.current_traversal,
                                          is_update,
                                          stack.id, rsrc_id)

    def _retrigger_new_traversal(self, cnxt, current_traversal, is_update,
                                 stack_id, rsrc_id):
        latest_stack = parser.Stack.load(cnxt, stack_id=stack_id,
                                         force_reload=True)
        if current_traversal != latest_stack.current_traversal:
            self.retrigger_check_resource(cnxt, rsrc_id, latest_stack)

    def _handle_stack_timeout(self, cnxt, stack):
        failure_reason = u'Timed out'
        stack.mark_failed(failure_reason)

    def _handle_resource_replacement(self, cnxt,
                                     current_traversal, new_tmpl_id, requires,
                                     rsrc, stack, adopt_stack_data):
        """Create a replacement resource and trigger a check on it."""
        try:
            new_res_id = rsrc.make_replacement(new_tmpl_id, requires)
        except exception.UpdateInProgress:
            LOG.info("No replacement created - "
                     "resource already locked by new traversal")
            return
        if new_res_id is None:
            LOG.info("No replacement created - "
                     "new traversal already in progress")
            self._retrigger_new_traversal(cnxt, current_traversal, True,
                                          stack.id, rsrc.id)
            return
        LOG.info("Replacing resource with new id %s", new_res_id)
        rpc_data = sync_point.serialize_input_data(self.input_data)
        self._rpc_client.check_resource(cnxt,
                                        new_res_id,
                                        current_traversal,
                                        rpc_data, True,
                                        adopt_stack_data)

    def _do_check_resource(self, cnxt, current_traversal, tmpl, resource_data,
                           is_update, rsrc, stack, adopt_stack_data):
        prev_template_id = rsrc.current_template_id
        try:
            if is_update:
                requires = set(d.primary_key for d in resource_data.values()
                               if d is not None)
                try:
                    check_resource_update(rsrc, tmpl.id, requires,
                                          self.engine_id,
                                          stack, self.msg_queue)
                except resource.UpdateReplace:
                    self._handle_resource_replacement(cnxt, current_traversal,
                                                      tmpl.id, requires,
                                                      rsrc, stack,
                                                      adopt_stack_data)
                    return False

            else:
                check_resource_cleanup(rsrc, tmpl.id, self.engine_id,
                                       stack.time_remaining(), self.msg_queue)

            return True
        except exception.UpdateInProgress:
            LOG.debug('Waiting for existing update to unlock resource %s',
                      rsrc.id)
            if self._stale_resource_needs_retry(cnxt, rsrc, prev_template_id):
                rpc_data = sync_point.serialize_input_data(self.input_data)
                self._rpc_client.check_resource(cnxt,
                                                rsrc.id,
                                                current_traversal,
                                                rpc_data, is_update,
                                                adopt_stack_data)
            else:
                rsrc.handle_preempt()
        except exception.ResourceFailure as ex:
            action = ex.action or rsrc.action
            reason = 'Resource %s failed: %s' % (action,
                                                 str(ex))
            self._handle_resource_failure(cnxt, is_update, rsrc.id,
                                          stack, reason)
        except scheduler.Timeout:
            self._handle_resource_failure(cnxt, is_update, rsrc.id,
                                          stack, u'Timed out')
        except CancelOperation:
            # Stack is already marked FAILED, so we just need to retrigger
            # in case a new traversal has started and is waiting on us.
            self._retrigger_new_traversal(cnxt, current_traversal, is_update,
                                          stack.id, rsrc.id)

        return False

    def retrigger_check_resource(self, cnxt, resource_id, stack):
        current_traversal = stack.current_traversal
        graph = stack.convergence_dependencies.graph()

        # When re-trigger received for latest traversal, first check if update
        # key is available in graph. If yes, the latest traversal is waiting
        # for update, otherwise it is waiting for delete. This is the case
        # regardless of which action (update or cleanup) from the previous
        # traversal was blocking it.
        update_key = parser.ConvergenceNode(resource_id, True)
        key = parser.ConvergenceNode(resource_id, update_key in graph)

        LOG.info('Re-trigger resource: %s', key)
        predecessors = set(graph[key])

        try:
            propagate_check_resource(cnxt, self._rpc_client, resource_id,
                                     current_traversal, predecessors, key,
                                     None, key.is_update, None)
        except exception.EntityNotFound as e:
            if e.entity != "Sync Point":
                raise

    def _initiate_propagate_resource(self, cnxt, resource_id,
                                     current_traversal, is_update, rsrc,
                                     stack):
        deps = stack.convergence_dependencies
        graph = deps.graph()
        graph_key = parser.ConvergenceNode(resource_id, is_update)

        if graph_key not in graph and rsrc.replaces is not None:
            # If we are a replacement, impersonate the replaced resource for
            # the purposes of calculating whether subsequent resources are
            # ready, since everybody has to work from the same version of the
            # graph. Our real resource ID is sent in the input_data, so the
            # dependencies will get updated to point to this resource in time
            # for the next traversal.
            graph_key = parser.ConvergenceNode(rsrc.replaces, is_update)

        def _get_input_data(req_node, input_forward_data=None):
            if req_node.is_update:
                if input_forward_data is None:
                    return rsrc.node_data().as_dict()
                else:
                    # do not re-resolve attrs
                    return input_forward_data
            else:
                # Don't send data if initiating clean-up for self i.e.
                # initiating delete of a replaced resource
                if req_node.rsrc_id != graph_key.rsrc_id:
                    # send replaced resource as needed_by if it exists
                    return (rsrc.replaced_by
                            if rsrc.replaced_by is not None
                            else resource_id)
            return None

        try:
            input_forward_data = None
            for req_node in sorted(deps.required_by(graph_key),
                                   key=lambda n: n.is_update):
                input_data = _get_input_data(req_node, input_forward_data)
                if req_node.is_update:
                    input_forward_data = input_data
                propagate_check_resource(
                    cnxt, self._rpc_client, req_node.rsrc_id,
                    current_traversal, set(graph[req_node]),
                    graph_key, input_data, req_node.is_update,
                    stack.adopt_stack_data)
            if is_update:
                if input_forward_data is None:
                    # we haven't resolved attribute data for the resource,
                    # so clear any old attributes so they may be re-resolved
                    rsrc.clear_stored_attributes()
                else:
                    rsrc.store_attributes()
            check_stack_complete(cnxt, stack, current_traversal,
                                 graph_key.rsrc_id, deps, graph_key.is_update)
        except exception.EntityNotFound as e:
            if e.entity == "Sync Point":
                # Reload the stack to determine the current traversal, and
                # check the SyncPoint for the current node to determine if
                # it is ready. If it is, then retrigger the current node
                # with the appropriate data for the latest traversal.
                stack = parser.Stack.load(cnxt, stack_id=rsrc.stack.id,
                                          force_reload=True)
                if current_traversal == stack.current_traversal:
                    LOG.debug('[%s] Traversal sync point missing.',
                              current_traversal)
                    return

                self.retrigger_check_resource(cnxt, resource_id, stack)
            else:
                raise

    def check(self, cnxt, resource_id, current_traversal,
              resource_data, is_update, adopt_stack_data,
              rsrc, stack):
        """Process a node in the dependency graph.

        The node may be associated with either an update or a cleanup of its
        associated resource.
        """
        if stack.has_timed_out():
            self._handle_stack_timeout(cnxt, stack)
            return

        tmpl = stack.t
        stack.adopt_stack_data = adopt_stack_data
        stack.thread_group_mgr = self.thread_group_mgr

        try:
            check_resource_done = self._do_check_resource(cnxt,
                                                          current_traversal,
                                                          tmpl, resource_data,
                                                          is_update,
                                                          rsrc, stack,
                                                          adopt_stack_data)

            if check_resource_done:
                # initiate check on next set of resources from graph
                self._initiate_propagate_resource(cnxt, resource_id,
                                                  current_traversal, is_update,
                                                  rsrc, stack)
        except BaseException as exc:
            with excutils.save_and_reraise_exception():
                msg = str(exc)
                LOG.exception("Unexpected exception in resource check.")
                self._handle_resource_failure(cnxt, is_update, rsrc.id,
                                              stack, msg)


def load_resource(cnxt, resource_id, resource_data,
                  current_traversal, is_update):
    try:
        return resource.Resource.load(cnxt, resource_id, current_traversal,
                                      is_update, resource_data)
    except (exception.ResourceNotFound, exception.NotFound):
        # can be ignored
        return None, None, None


def check_stack_complete(cnxt, stack, current_traversal, sender_id, deps,
                         is_update):
    """Mark the stack complete if the update is complete.

    Complete is currently in the sense that all desired resources are in
    service, not that superfluous ones have been cleaned up.
    """
    roots = set(deps.roots())

    if (sender_id, is_update) not in roots:
        return

    def mark_complete(stack_id, data):
        stack.mark_complete()

    sender_key = parser.ConvergenceNode(sender_id, is_update)
    sync_point.sync(cnxt, stack.id, current_traversal, True,
                    mark_complete, roots, {sender_key: None})


def propagate_check_resource(cnxt, rpc_client, next_res_id,
                             current_traversal, predecessors, sender_key,
                             sender_data, is_update, adopt_stack_data):
    """Trigger processing of node if all of its dependencies are satisfied."""
    def do_check(entity_id, data):
        rpc_client.check_resource(cnxt, entity_id, current_traversal,
                                  data, is_update, adopt_stack_data)

    sync_point.sync(cnxt, next_res_id, current_traversal,
                    is_update, do_check, predecessors,
                    {sender_key: sender_data})


def _check_for_message(msg_queue):
    if msg_queue is None:
        return
    try:
        message = msg_queue.get_nowait()
    except eventlet.queue.Empty:
        return

    if message == rpc_api.THREAD_CANCEL:
        raise CancelOperation

    LOG.error('Unknown message "%s" received', message)


def check_resource_update(rsrc, template_id, requires, engine_id,
                          stack, msg_queue):
    """Create or update the Resource if appropriate."""
    check_message = functools.partial(_check_for_message, msg_queue)
    if rsrc.action == resource.Resource.INIT:
        rsrc.create_convergence(template_id, requires, engine_id,
                                stack.time_remaining(), check_message)
    else:
        rsrc.update_convergence(template_id, requires, engine_id,
                                stack.time_remaining(), stack,
                                check_message)


def check_resource_cleanup(rsrc, template_id, engine_id,
                           timeout, msg_queue):
    """Delete the Resource if appropriate."""
    check_message = functools.partial(_check_for_message, msg_queue)
    rsrc.delete_convergence(template_id, engine_id, timeout,
                            check_message)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from heat.common import exception
from heat.common import template_format
from heat.engine.cfn import functions as cfn_funcs
from heat.engine import function
from heat.engine.hot import functions as hot_funcs
from heat.engine import properties
from heat.engine import rsrc_defn
from heat.tests import common
from heat.tests import utils

TEMPLATE_WITH_EX_REF_IMPLICIT_DEPEND = '''
heat_template_version: 2016-10-14
resources:
  test1:
    type: OS::Heat::TestResource
    external_id: foobar
    properties:
        value: {get_resource: test2}
  test2:
    type: OS::Heat::TestResource
'''

TEMPLATE_WITH_INVALID_EXPLICIT_DEPEND = '''
heat_template_version: 2016-10-14
resources:
  test1:
    type: OS::Heat::TestResource
  test3:
    type: OS::Heat::TestResource
    depends_on: test2
'''


class ResourceDefinitionTest(common.HeatTestCase):

    def make_me_one_with_everything(self):
        return rsrc_defn.ResourceDefinition(
            'rsrc', 'SomeType',
            properties={'Foo': cfn_funcs.Join(None,
                                              'Fn::Join',
                                              ['a', ['b', 'r']]),
                        'Blarg': 'wibble'},
            metadata={'Baz': cfn_funcs.Join(None,
                                            'Fn::Join',
                                            ['u', ['q', '', 'x']])},
            depends=['other_resource'],
            deletion_policy='Retain',
            update_policy={'SomePolicy': {}})

    def test_properties_default(self):
        rd = rsrc_defn.ResourceDefinition('rsrc', 'SomeType')
        self.assertEqual({}, rd.properties({}))

    def test_properties(self):
        rd = self.make_me_one_with_everything()

        schema = {
            'Foo': properties.Schema(properties.Schema.STRING),
            'Blarg': properties.Schema(properties.Schema.STRING, default=''),
            'Baz': properties.Schema(properties.Schema.STRING, default='quux'),
        }

        props = rd.properties(schema)
        self.assertEqual('bar', props['Foo'])
        self.assertEqual('wibble', props['Blarg'])
        self.assertEqual('quux', props['Baz'])

    def test_metadata_default(self):
        rd = rsrc_defn.ResourceDefinition('rsrc', 'SomeType')
        self.assertEqual({}, rd.metadata())

    def test_metadata(self):
        rd = self.make_me_one_with_everything()
        metadata = rd.metadata()
        self.assertEqual({'Baz': 'quux'}, metadata)
        self.assertIsInstance(metadata['Baz'], str)

    def test_dependencies_default(self):
        rd = rsrc_defn.ResourceDefinition('rsrc', 'SomeType')
        stack = {'foo': 'FOO', 'bar': 'BAR'}
        self.assertEqual(set(), rd.required_resource_names())
        self.assertEqual([], list(rd.dependencies(stack)))

    def test_dependencies_explicit(self):
        rd = rsrc_defn.ResourceDefinition('rsrc', 'SomeType', depends=['foo'])
        stack = {'foo': 'FOO', 'bar': 'BAR'}
        self.assertEqual({'foo'}, rd.required_resource_names())
        self.assertEqual(['FOO'], list(rd.dependencies(stack)))

    def test_dependencies_explicit_ext(self):
        rd = rsrc_defn.ResourceDefinition('rsrc', 'SomeType', depends=['foo'],
                                          external_id='abc')
        stack = {'foo': 'FOO', 'bar': 'BAR'}
        self.assertRaises(
            exception.InvalidExternalResourceDependency,
            rd.dependencies, stack)

    def test_dependencies_implicit_ext(self):
        t = template_format.parse(TEMPLATE_WITH_EX_REF_IMPLICIT_DEPEND)
        stack = utils.parse_stack(t)
        rsrc = stack['test1']
        self.assertEqual([], list(rsrc.t.dependencies(stack)))

    def test_dependencies_explicit_invalid(self):
        t = template_format.parse(TEMPLATE_WITH_INVALID_EXPLICIT_DEPEND)
        stack = utils.parse_stack(t)
        rd = stack.t.resource_definitions(stack)['test3']
        self.assertEqual({'test2'}, rd.required_resource_names())
        self.assertRaises(exception.InvalidTemplateReference,
                          lambda: list(rd.dependencies(stack)))

    def test_deletion_policy_default(self):
        rd = rsrc_defn.ResourceDefinition('rsrc', 'SomeType')
        self.assertEqual(rsrc_defn.ResourceDefinition.DELETE,
                         rd.deletion_policy())

    def test_deletion_policy(self):
        for policy in rsrc_defn.ResourceDefinition.DELETION_POLICIES:
            rd = rsrc_defn.ResourceDefinition('rsrc', 'SomeType',
                                              deletion_policy=policy)
            self.assertEqual(policy, rd.deletion_policy())

    def test_deletion_policy_invalid(self):
        self.assertRaises(AssertionError,
                          rsrc_defn.ResourceDefinition,
                          'rsrc', 'SomeType', deletion_policy='foo')

    def test_update_policy_default(self):
        rd = rsrc_defn.ResourceDefinition('rsrc', 'SomeType')
        self.assertEqual({}, rd.update_policy({}))

    def test_update_policy(self):
        rd = self.make_me_one_with_everything()

        policy_schema = {'Foo': properties.Schema(properties.Schema.STRING,
                                                  default='bar')}
        schema = {
            'SomePolicy': properties.Schema(properties.Schema.MAP,
                                            schema=policy_schema),
        }

        up = rd.update_policy(schema)
        self.assertEqual('bar', up['SomePolicy']['Foo'])

    def test_freeze(self):
        rd = self.make_me_one_with_everything()

        frozen = rd.freeze()
        self.assertEqual('bar', frozen._properties['Foo'])
        self.assertEqual('quux', frozen._metadata['Baz'])

    def test_freeze_override(self):
        rd = self.make_me_one_with_everything()

        frozen = rd.freeze(metadata={'Baz': 'wibble'})
        self.assertEqual('bar', frozen._properties['Foo'])
        self.assertEqual('wibble', frozen._metadata['Baz'])

    def test_freeze_nullable_top_level(self):
        class NullFunction(function.Function):
            def result(self):
                return Ellipsis

        null_func = NullFunction(None, 'null', [])

        rd = rsrc_defn.ResourceDefinition(
            'rsrc', 'SomeType',
            properties=null_func,
            metadata=null_func,
            update_policy=null_func)

        frozen = rd.freeze()
        self.assertIsNone(frozen._properties)
        self.assertIsNone(frozen._metadata)
        self.assertIsNone(frozen._update_policy)

        rd2 = rsrc_defn.ResourceDefinition(
            'rsrc', 'SomeType',
            properties={'Foo': null_func,
                        'Blarg': 'wibble'},
            metadata={'Bar': null_func,
                      'Baz': 'quux'},
            update_policy={'some_policy': null_func})

        frozen2 = rd2.freeze()
        self.assertNotIn('Foo', frozen2._properties)
        self.assertEqual('wibble', frozen2._properties['Blarg'])
        self.assertNotIn('Bar', frozen2._metadata)
        self.assertEqual('quux', frozen2._metadata['Baz'])
        self.assertEqual({}, frozen2._update_policy)

    def test_render_hot(self):
        rd = self.make_me_one_with_everything()

        expected_hot = {
            'type': 'SomeType',
            'properties': {'Foo': {'Fn::Join': ['a', ['b', 'r']]},
                           'Blarg': 'wibble'},
            'metadata': {'Baz': {'Fn::Join': ['u', ['q', '', 'x']]}},
            'depends_on': ['other_resource'],
            'deletion_policy': 'Retain',
            'update_policy': {'SomePolicy': {}},
        }

        self.assertEqual(expected_hot, rd.render_hot())

    def test_render_hot_empty(self):
        rd = rsrc_defn.ResourceDefinition('rsrc', 'SomeType')

        expected_hot = {
            'type': 'SomeType',
        }

        self.assertEqual(expected_hot, rd.render_hot())

    def test_template_equality(self):
        class FakeStack(object):
            def __init__(self, params):
                self.parameters = params

        def get_param_defn(value):
            stack = FakeStack({'Foo': value})
            param_func = hot_funcs.GetParam(stack, 'get_param', 'Foo')

            return rsrc_defn.ResourceDefinition('rsrc', 'SomeType',
                                                {'Foo': param_func})

        self.assertEqual(get_param_defn('bar'), get_param_defn('baz'))

    def test_hash_equal(self):
        rd1 = self.make_me_one_with_everything()
        rd2 = self.make_me_one_with_everything()
        self.assertEqual(rd1, rd2)
        self.assertEqual(hash(rd1), hash(rd2))

    def test_hash_names(self):
        rd1 = rsrc_defn.ResourceDefinition('rsrc1', 'SomeType')
        rd2 = rsrc_defn.ResourceDefinition('rsrc2', 'SomeType')
        self.assertEqual(rd1, rd2)
        self.assertEqual(hash(rd1), hash(rd2))

    def test_hash_types(self):
        rd1 = rsrc_defn.ResourceDefinition('rsrc', 'SomeType1')
        rd2 = rsrc_defn.ResourceDefinition('rsrc', 'SomeType2')
        self.assertNotEqual(rd1, rd2)
        self.assertNotEqual(hash(rd1), hash(rd2))


class ResourceDefinitionDiffTest(common.HeatTestCase):
    def test_properties_diff(self):
        before = rsrc_defn.ResourceDefinition('rsrc', 'SomeType',
                                              properties={'Foo': 'blarg'},
                                              update_policy={'baz': 'quux'},
                                              metadata={'baz': 'quux'})
        after = rsrc_defn.ResourceDefinition('rsrc', 'SomeType',
                                             properties={'Foo': 'wibble'},
                                             update_policy={'baz': 'quux'},
                                             metadata={'baz': 'quux'})

        diff = after - before
        self.assertTrue(diff.properties_changed())
        self.assertFalse(diff.update_policy_changed())
        self.assertFalse(diff.metadata_changed())
        self.assertTrue(diff)

    def test_update_policy_diff(self):
        before = rsrc_defn.ResourceDefinition('rsrc', 'SomeType',
                                              properties={'baz': 'quux'},
                                              update_policy={'Foo': 'blarg'},
                                              metadata={'baz': 'quux'})
        after = rsrc_defn.ResourceDefinition('rsrc', 'SomeType',
                                             properties={'baz': 'quux'},
                                             update_policy={'Foo': 'wibble'},
                                             metadata={'baz': 'quux'})

        diff = after - before
        self.assertFalse(diff.properties_changed())
        self.assertTrue(diff.update_policy_changed())
        self.assertFalse(diff.metadata_changed())
        self.assertTrue(diff)

    def test_metadata_diff(self):
        before = rsrc_defn.ResourceDefinition('rsrc', 'SomeType',
                                              properties={'baz': 'quux'},
                                              update_policy={'baz': 'quux'},
                                              metadata={'Foo': 'blarg'})
        after = rsrc_defn.ResourceDefinition('rsrc', 'SomeType',
                                             properties={'baz': 'quux'},
                                             update_policy={'baz': 'quux'},
                                             metadata={'Foo': 'wibble'})

        diff = after - before
        self.assertFalse(diff.properties_changed())
        self.assertFalse(diff.update_policy_changed())
        self.assertTrue(diff.metadata_changed())
        self.assertTrue(diff)

    def test_no_diff(self):
        before = rsrc_defn.ResourceDefinition('rsrc', 'SomeType',
                                              properties={'Foo': 'blarg'},
                                              update_policy={'bar': 'quux'},
                                              metadata={'baz': 'wibble'},
                                              depends=['other_resource'],
                                              deletion_policy='Delete')
        after = rsrc_defn.ResourceDefinition('rsrc', 'SomeType',
                                             properties={'Foo': 'blarg'},
                                             update_policy={'bar': 'quux'},
                                             metadata={'baz': 'wibble'},
                                             depends=['other_other_resource'],
                                             deletion_policy='Retain')

        diff = after - before
        self.assertFalse(diff.properties_changed())
        self.assertFalse(diff.update_policy_changed())
        self.assertFalse(diff.metadata_changed())
        self.assertFalse(diff)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from oslo_config import cfg
from zunclient import exceptions as zc_exc

from heat.common import exception
from heat.common import template_format
from heat.engine.clients.os import neutron
from heat.engine.clients.os import zun
from heat.engine.resources.openstack.zun import container
from heat.engine import scheduler
from heat.engine import template
from heat.tests import common
from heat.tests import utils

zun_template = '''
heat_template_version: 2017-09-01

resources:
  test_container:
    type: OS::Zun::Container
    properties:
      name: test_container
      image: "cirros:latest"
      command: sleep 10000
      cpu: 0.1
      memory: 100
      environment:
        myenv: foo
      workdir: /testdir
      labels:
        mylabel: bar
      image_pull_policy: always
      restart_policy: on-failure:2
      interactive: false
      tty: false
      image_driver: docker
      hints:
        hintkey: hintval
      hostname: myhost
      security_groups:
        - my_seg
      mounts:
        - volume_size: 1
          mount_path: /data
        - volume_id: 6ec29ba3-bf2c-4276-a88e-3670ea5abc80
          mount_path: /data2
      networks:
        - network: mynet
          fixed_ip: 10.0.0.4
        - network: mynet2
          fixed_ip: fe80::3
        - port: myport
'''

zun_template_minimum = '''
heat_template_version: 2017-09-01

resources:
  test_container:
    type: OS::Zun::Container
    properties:
      name: test_container
      image: "cirros:latest"
'''


def create_fake_iface(port=None, net=None, mac=None, ip=None, subnet=None):
    class fake_interface(object):
        def __init__(self, port_id, net_id, mac_addr, fixed_ip, subnet_id):
            self.port_id = port_id
            self.net_id = net_id
            self.mac_addr = mac_addr
            self.fixed_ips = [{'ip_address': fixed_ip, 'subnet_id': subnet_id}]

    return fake_interface(port, net, mac, ip, subnet)


class ZunContainerTest(common.HeatTestCase):

    def setUp(self):
        super(ZunContainerTest, self).setUp()

        self.resource_id = '12345'
        self.fake_name = 'test_container'
        self.fake_image = 'cirros:latest'
        self.fake_command = 'sleep 10000'
        self.fake_cpu = 0.1
        self.fake_memory = 100
        self.fake_env = {'myenv': 'foo'}
        self.fake_workdir = '/testdir'
        self.fake_labels = {'mylabel': 'bar'}
        self.fake_image_policy = 'always'
        self.fake_restart_policy = {'MaximumRetryCount': '2',
                                    'Name': 'on-failure'}
        self.fake_interactive = False
        self.fake_tty = False
        self.fake_image_driver = 'docker'
        self.fake_hints = {'hintkey': 'hintval'}
        self.fake_hostname = 'myhost'
        self.fake_security_groups = ['my_seg']
        self.fake_mounts = [
            {'volume_id': None, 'volume_size': 1, 'mount_path': '/data'},
            {'volume_id': '6ec29ba3-bf2c-4276-a88e-3670ea5abc80',
             'volume_size': None, 'mount_path': '/data2'}]
        self.fake_mounts_args = [
            {'size': 1, 'destination': '/data'},
            {'source': '6ec29ba3-bf2c-4276-a88e-3670ea5abc80',
             'destination': '/data2'}]
        self.fake_networks = [
            {'network': 'mynet', 'port': None, 'fixed_ip': '10.0.0.4'},
            {'network': 'mynet2', 'port': None, 'fixed_ip': 'fe80::3'},
            {'network': None, 'port': 'myport', 'fixed_ip': None}]
        self.fake_networks_args = [
            {'network': 'mynet', 'v4-fixed-ip': '10.0.0.4'},
            {'network': 'mynet2', 'v6-fixed-ip': 'fe80::3'},
            {'port': 'myport'}]

        self.fake_network_id = '9c11d847-99ce-4a83-82da-9827362a68e8'
        self.fake_network_name = 'private'
        self.fake_networks_attr = {
            'networks': [
                {
                    'id': self.fake_network_id,
                    'name': self.fake_network_name,
                }
            ]
        }
        self.fake_address = {
            'version': 4,
            'addr': '10.0.0.12',
            'port': 'ab5c12d8-f414-48a3-b765-8ce34a6714d2'
        }
        self.fake_addresses = {
            self.fake_network_id: [self.fake_address]
        }
        self.fake_extended_addresses = {
            self.fake_network_id: [self.fake_address],
            self.fake_network_name: [self.fake_address],
        }

        t = template_format.parse(zun_template)
        self.stack = utils.parse_stack(t)
        resource_defns = self.stack.t.resource_definitions(self.stack)
        self.rsrc_defn = resource_defns[self.fake_name]
        self.client = mock.Mock()
        self.patchobject(container.Container, 'client',
                         return_value=self.client)
        self.neutron_client = mock.Mock()
        self.patchobject(container.Container, 'neutron',
                         return_value=self.neutron_client)
        self.stub_VolumeConstraint_validate()
        self.mock_update = self.patchobject(zun.ZunClientPlugin,
                                            'update_container')
        self.stub_PortConstraint_validate()
        self.mock_find = self.patchobject(
            neutron.NeutronClientPlugin,
            'find_resourceid_by_name_or_id',
            side_effect=lambda x, y: y)
        self.mock_attach = self.patchobject(zun.ZunClientPlugin,
                                            'network_attach')
        self.mock_detach = self.patchobject(zun.ZunClientPlugin,
                                            'network_detach')
        self.mock_attach_check = self.patchobject(zun.ZunClientPlugin,
                                                  'check_network_attach',
                                                  return_value=True)
        self.mock_detach_check = self.patchobject(zun.ZunClientPlugin,
                                                  'check_network_detach',
                                                  return_value=True)

    def _mock_get_client(self):
        value = mock.MagicMock()
        value.name = self.fake_name
        value.image = self.fake_image
        value.command = self.fake_command
        value.cpu = self.fake_cpu
        value.memory = self.fake_memory
        value.environment = self.fake_env
        value.workdir = self.fake_workdir
        value.labels = self.fake_labels
        value.image_pull_policy = self.fake_image_policy
        value.restart_policy = self.fake_restart_policy
        value.interactive = self.fake_interactive
        value.tty = self.fake_tty
        value.image_driver = self.fake_image_driver
        value.hints = self.fake_hints
        value.hostname = self.fake_hostname
        value.security_groups = self.fake_security_groups
        value.addresses = self.fake_addresses
        value.to_dict.return_value = value.__dict__

        self.client.containers.get.return_value = value

    def _create_resource(self, name, snippet, stack, status='Running'):
        value = mock.MagicMock(uuid=self.resource_id)
        self.client.containers.run.return_value = value
        get_rv = mock.MagicMock(status=status)
        self.client.containers.get.return_value = get_rv
        c = container.Container(name, snippet, stack)
        return c

    def test_create(self):
        c = self._create_resource('container', self.rsrc_defn,
                                  self.stack)
        # validate the properties
        self.assertEqual(
            self.fake_name,
            c.properties.get(container.Container.NAME))
        self.assertEqual(
            self.fake_image,
            c.properties.get(container.Container.IMAGE))
        self.assertEqual(
            self.fake_command,
            c.properties.get(container.Container.COMMAND))
        self.assertEqual(
            self.fake_cpu,
            c.properties.get(container.Container.CPU))
        self.assertEqual(
            self.fake_memory,
            c.properties.get(container.Container.MEMORY))
        self.assertEqual(
            self.fake_env,
            c.properties.get(container.Container.ENVIRONMENT))
        self.assertEqual(
            self.fake_workdir,
            c.properties.get(container.Container.WORKDIR))
        self.assertEqual(
            self.fake_labels,
            c.properties.get(container.Container.LABELS))
        self.assertEqual(
            self.fake_image_policy,
            c.properties.get(container.Container.IMAGE_PULL_POLICY))
        self.assertEqual(
            'on-failure:2',
            c.properties.get(container.Container.RESTART_POLICY))
        self.assertEqual(
            self.fake_interactive,
            c.properties.get(container.Container.INTERACTIVE))
        self.assertEqual(
            self.fake_tty,
            c.properties.get(container.Container.TTY))
        self.assertEqual(
            self.fake_image_driver,
            c.properties.get(container.Container.IMAGE_DRIVER))
        self.assertEqual(
            self.fake_hints,
            c.properties.get(container.Container.HINTS))
        self.assertEqual(
            self.fake_hostname,
            c.properties.get(container.Container.HOSTNAME))
        self.assertEqual(
            self.fake_security_groups,
            c.properties.get(container.Container.SECURITY_GROUPS))
        self.assertEqual(
            self.fake_mounts,
            c.properties.get(container.Container.MOUNTS))
        self.assertEqual(
            self.fake_networks,
            c.properties.get(container.Container.NETWORKS))

        scheduler.TaskRunner(c.create)()
        self.assertEqual(self.resource_id, c.resource_id)
        self.assertEqual((c.CREATE, c.COMPLETE), c.state)
        self.assertEqual('containers', c.entity)
        self.client.containers.run.assert_called_once_with(
            name=self.fake_name,
            image=self.fake_image,
            command=self.fake_command.split(),
            cpu=self.fake_cpu,
            memory=self.fake_memory,
            environment=self.fake_env,
            workdir=self.fake_workdir,
            labels=self.fake_labels,
            image_pull_policy=self.fake_image_policy,
            restart_policy=self.fake_restart_policy,
            interactive=self.fake_interactive,
            tty=self.fake_tty,
            image_driver=self.fake_image_driver,
            hints=self.fake_hints,
            hostname=self.fake_hostname,
            security_groups=self.fake_security_groups,
            mounts=self.fake_mounts_args,
            nets=self.fake_networks_args,
        )

    def test_container_create_failed(self):
        cfg.CONF.set_override('action_retry_limit', 0)
        c = self._create_resource('container', self.rsrc_defn, self.stack,
                                  status='Error')
        exc = self.assertRaises(
            exception.ResourceFailure,
            scheduler.TaskRunner(c.create))
        self.assertEqual((c.CREATE, c.FAILED), c.state)
        self.assertIn("Error in creating container ", str(exc))

    def test_container_create_unknown_status(self):
        c = self._create_resource('container', self.rsrc_defn, self.stack,
                                  status='FOO')
        exc = self.assertRaises(
            exception.ResourceFailure,
            scheduler.TaskRunner(c.create))
        self.assertEqual((c.CREATE, c.FAILED), c.state)
        self.assertIn("Unknown status Container", str(exc))

    def test_container_update(self):
        c = self._create_resource('container', self.rsrc_defn, self.stack)
        scheduler.TaskRunner(c.create)()
        t = template_format.parse(zun_template)
        new_t = copy.deepcopy(t)
        new_t['resources'][self.fake_name]['properties']['name'] = \
            'fake-container'
        new_t['resources'][self.fake_name]['properties']['cpu'] = 10
        new_t['resources'][self.fake_name]['properties']['memory'] = 10
        rsrc_defns = template.Template(new_t).resource_definitions(self.stack)
        new_c = rsrc_defns[self.fake_name]
        scheduler.TaskRunner(c.update, new_c)()
        self.mock_update.assert_called_once_with(
            self.resource_id,
            cpu=10, memory=10, name='fake-container')
        self.assertEqual((c.UPDATE, c.COMPLETE), c.state)

    def _test_container_update_None_networks(self, new_networks):
        t = template_format.parse(zun_template_minimum)
        stack = utils.parse_stack(t)
        resource_defns = stack.t.resource_definitions(stack)
        rsrc_defn = resource_defns[self.fake_name]
        c = self._create_resource('container', rsrc_defn, stack)
        scheduler.TaskRunner(c.create)()

        new_t = copy.deepcopy(t)
        new_t['resources'][self.fake_name]['properties']['networks'] = \
            new_networks
        rsrc_defns = template.Template(new_t).resource_definitions(stack)
        new_c = rsrc_defns[self.fake_name]
        iface = create_fake_iface(
            port='aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            net='450abbc9-9b6d-4d6f-8c3a-c47ac34100ef',
            ip='1.2.3.4')
        self.client.containers.network_list.return_value = [iface]
        scheduler.TaskRunner(c.update, new_c)()
        self.assertEqual((c.UPDATE, c.COMPLETE), c.state)
        self.client.containers.network_list.assert_called_once_with(
            self.resource_id)

    def test_container_update_None_networks_with_port(self):
        new_networks = [{'port': '2a60cbaa-3d33-4af6-a9ce-83594ac546fc'}]
        self._test_container_update_None_networks(new_networks)
        self.assertEqual(1, self.mock_attach.call_count)
        self.assertEqual(1, self.mock_detach.call_count)
        self.assertEqual(1, self.mock_attach_check.call_count)
        self.assertEqual(1, self.mock_detach_check.call_count)

    def test_container_update_None_networks_with_network_id(self):
        new_networks = [{'network': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                         'fixed_ip': '1.2.3.4'}]
        self._test_container_update_None_networks(new_networks)
        self.assertEqual(1, self.mock_attach.call_count)
        self.assertEqual(1, self.mock_detach.call_count)
        self.assertEqual(1, self.mock_attach_check.call_count)
        self.assertEqual(1, self.mock_detach_check.call_count)

    def test_container_update_None_networks_with_complex_parameters(self):
        new_networks = [{'network': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                         'fixed_ip': '1.2.3.4',
                         'port': '2a60cbaa-3d33-4af6-a9ce-83594ac546fc'}]
        self._test_container_update_None_networks(new_networks)
        self.assertEqual(1, self.mock_attach.call_count)
        self.assertEqual(1, self.mock_detach.call_count)
        self.assertEqual(1, self.mock_attach_check.call_count)
        self.assertEqual(1, self.mock_detach_check.call_count)

    def test_server_update_empty_networks_to_None(self):
        new_networks = None
        self._test_container_update_None_networks(new_networks)
        self.assertEqual(0, self.mock_attach.call_count)
        self.assertEqual(0, self.mock_detach.call_count)
        self.assertEqual(0, self.mock_attach_check.call_count)
        self.assertEqual(0, self.mock_detach_check.call_count)

    def _test_container_update_networks(self, new_networks):
        c = self._create_resource('container', self.rsrc_defn, self.stack)
        scheduler.TaskRunner(c.create)()
        t = template_format.parse(zun_template)
        new_t = copy.deepcopy(t)
        new_t['resources'][self.fake_name]['properties']['networks'] = \
            new_networks
        rsrc_defns = template.Template(new_t).resource_definitions(self.stack)
        new_c = rsrc_defns[self.fake_name]
        sec_uuids = ['86c0f8ae-23a8-464f-8603-c54113ef5467']
        self.patchobject(neutron.NeutronClientPlugin,
                         'get_secgroup_uuids', return_value=sec_uuids)
        ifaces = [
            create_fake_iface(port='95e25541-d26a-478d-8f36-ae1c8f6b74dc',
                              net='mynet',
                              ip='10.0.0.4'),
            create_fake_iface(port='450abbc9-9b6d-4d6f-8c3a-c47ac34100ef',
                              net='mynet2',
                              ip='fe80::3'),
            create_fake_iface(port='myport',
                              net='aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                              ip='21.22.23.24')]
        self.client.containers.network_list.return_value = ifaces
        scheduler.TaskRunner(c.update, new_c)()
        self.assertEqual((c.UPDATE, c.COMPLETE), c.state)
        self.client.containers.network_list.assert_called_once_with(
            self.resource_id)

    def test_container_update_networks_with_complex_parameters(self):
        new_networks = [
            {'network': 'mynet',
             'fixed_ip': '10.0.0.4'},
            {'port': '2a60cbaa-3d33-4af6-a9ce-83594ac546fc'}]
        self._test_container_update_networks(new_networks)
        self.assertEqual(2, self.mock_detach.call_count)
        self.assertEqual(1, self.mock_attach.call_count)
        self.assertEqual(2, self.mock_detach_check.call_count)
        self.assertEqual(1, self.mock_attach_check.call_count)

    def test_container_update_networks_with_None(self):
        new_networks = None
        self._test_container_update_networks(new_networks)
        self.assertEqual(3, self.mock_detach.call_count)
        self.assertEqual(1, self.mock_attach.call_count)
        self.assertEqual(3, self.mock_detach_check.call_count)
        self.assertEqual(1, self.mock_attach_check.call_count)

    def test_container_update_old_networks_to_empty_list(self):
        new_networks = []
        self._test_container_update_networks(new_networks)
        self.assertEqual(3, self.mock_detach.call_count)
        self.assertEqual(1, self.mock_attach.call_count)
        self.assertEqual(3, self.mock_detach_check.call_count)
        self.assertEqual(1, self.mock_attach_check.call_count)

    def test_container_update_remove_network_non_empty(self):
        new_networks = [
            {'network': 'mynet',
             'fixed_ip': '10.0.0.4'},
            {'port': 'myport'}]
        self._test_container_update_networks(new_networks)
        self.assertEqual(1, self.mock_detach.call_count)
        self.assertEqual(0, self.mock_attach.call_count)
        self.assertEqual(1, self.mock_detach_check.call_count)
        self.assertEqual(0, self.mock_attach_check.call_count)

    def test_container_delete(self):
        c = self._create_resource('container', self.rsrc_defn, self.stack)
        scheduler.TaskRunner(c.create)()
        self.patchobject(self.client.containers, 'get',
                         side_effect=[c, zc_exc.NotFound('Not Found')])
        scheduler.TaskRunner(c.delete)()
        self.assertEqual((c.DELETE, c.COMPLETE), c.state)
        self.client.containers.delete.assert_called_once_with(
            c.resource_id, stop=True)

    def test_container_delete_not_found(self):
        c = self._create_resource('container', self.rsrc_defn, self.stack)
        scheduler.TaskRunner(c.create)()
        c.client_plugin = mock.MagicMock()
        self.client.containers.delete.side_effect = Exception('Not Found')
        scheduler.TaskRunner(c.delete)()
        self.assertEqual((c.DELETE, c.COMPLETE), c.state)
        self.client.containers.delete.assert_called_once_with(
            c.resource_id, stop=True)
        mock_ignore_not_found = c.client_plugin.return_value.ignore_not_found
        self.assertEqual(1, mock_ignore_not_found.call_count)

    def test_container_get_live_state(self):
        c = self._create_resource('container', self.rsrc_defn, self.stack)
        scheduler.TaskRunner(c.create)()
        self._mock_get_client()
        reality = c.get_live_state(c.properties)
        self.assertEqual(
            {
                container.Container.NAME: self.fake_name,
                container.Container.CPU: self.fake_cpu,
                container.Container.MEMORY: self.fake_memory,
            }, reality)

    def test_resolve_attributes(self):
        self.neutron_client.list_networks.return_value = \
            self.fake_networks_attr
        c = self._create_resource('container', self.rsrc_defn, self.stack)
        scheduler.TaskRunner(c.create)()
        self._mock_get_client()
        self.assertEqual(
            self.fake_name,
            c._resolve_attribute(container.Container.NAME))
        self.assertEqual(
            self.fake_extended_addresses,
            c._resolve_attribute(container.Container.ADDRESSES))

    def test_resolve_attributes_duplicate_net_name(self):
        self.neutron_client.list_networks.return_value = {
            'networks': [
                {'id': 'fake_net_id', 'name': 'test'},
                {'id': 'fake_net_id2', 'name': 'test'},
            ]
        }
        self.fake_addresses = {
            'fake_net_id': [{'addr': '10.0.0.12'}],
            'fake_net_id2': [{'addr': '10.100.0.12'}],
        }
        self.fake_extended_addresses = {
            'fake_net_id': [{'addr': '10.0.0.12'}],
            'fake_net_id2': [{'addr': '10.100.0.12'}],
            'test': [{'addr': '10.0.0.12'}, {'addr': '10.100.0.12'}],
        }
        c = self._create_resource('container', self.rsrc_defn, self.stack)
        scheduler.TaskRunner(c.create)()
        self._mock_get_client()
        self._assert_addresses(
            self.fake_extended_addresses,
            c._resolve_attribute(container.Container.ADDRESSES))

    def _assert_addresses(self, expected, actual):
        matched = True
        if len(expected) != len(actual):
            matched = False
        for key in expected:
            if key not in actual:
                matched = False
                break
            list1 = expected[key]
            list1 = sorted(list1, key=lambda x: sorted(x.values()))
            list2 = actual[key]
            list2 = sorted(list2, key=lambda x: sorted(x.values()))
            if list1 != list2:
                matched = False
                break

        if not matched:
            raise AssertionError(
                'Addresses is unmatched:\n reference = ' + str(expected) +
                '\nactual = ' + str(actual))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import strutils

from heat.common.i18n import _


def extract_bool(name, value):
    """Convert any true/false string to its corresponding boolean value.

    Value is case insensitive.
    """
    if str(value).lower() not in ('true', 'false'):
        raise ValueError(_('Unrecognized value "%(value)s" for "%(name)s", '
                           'acceptable values are: true, false.')
                         % {'value': value, 'name': name})
    return strutils.bool_from_string(value, strict=True)


def delim_string_to_list(value):
    if value is None:
        return None

    if value == '':
        return []

    return value.split(',')


def extract_int(name, value, allow_zero=True, allow_negative=False):
    if value is None:
        return None

    if not strutils.is_int_like(value):
        raise ValueError(_("Only integer is acceptable by "
                           "'%(name)s'.") % {'name': name})

    if value in ('0', 0):
        if allow_zero:
            return int(value)
        raise ValueError(_("Only non-zero integer is acceptable by "
                           "'%(name)s'.") % {'name': name})
    try:
        result = int(value)
    except (TypeError, ValueError):
        raise ValueError(_("Value '%(value)s' is invalid for '%(name)s' "
                           "which only accepts integer.") %
                         {'name': name, 'value': value})

    if allow_negative is False and result < 0:
        raise ValueError(_("Value '%(value)s' is invalid for '%(name)s' "
                           "which only accepts non-negative integer.") %
                         {'name': name, 'value': value})

    return result


def extract_tags(subject):
    tags = subject.split(',')
    for tag in tags:
        if len(tag) > 80:
            raise ValueError(_('Invalid tag, "%s" is longer than 80 '
                               'characters') % tag)
    return tags


def extract_template_type(subject):
    template_type = subject.lower()
    if template_type not in ('cfn', 'hot'):
        raise ValueError(_('Invalid template type "%(value)s", valid '
                           'types are: cfn, hot.') %
                         {'value': subject})
    return template_type



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Controller that returns information on the heat API versions."""

import http.client

from oslo_serialization import jsonutils
import webob.dec


class Controller(object):

    """A controller that produces information on the heat API versions."""

    def __init__(self, conf):
        self.conf = conf

    @webob.dec.wsgify
    def __call__(self, req):
        """Respond to a request for all OpenStack API versions."""
        version_objs = [
            {
                "id": "v1.0",
                "status": "CURRENT",
                "links": [
                    {
                        "rel": "self",
                        "href": self.get_href(req)
                    }]
            }]

        body = jsonutils.dumps(dict(versions=version_objs))

        response = webob.Response(request=req,
                                  status=http.client.MULTIPLE_CHOICES,
                                  content_type='application/json')
        # NOTE(pas-ha) in WebOb, Response.body accepts only bytes,
        # and Response.text accepts only unicode.
        response.text = str(body)

        return response

    def get_href(self, req):
        return "%s/v1/" % req.application_url



# Copyright 2014 Intel Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""SoftwareConfig object."""

from oslo_versionedobjects import base
from oslo_versionedobjects import fields

from heat.db.sqlalchemy import api as db_api
from heat.objects import base as heat_base
from heat.objects import fields as heat_fields


class SoftwareConfig(
        heat_base.HeatObject,
        base.VersionedObjectDictCompat,
        base.ComparableVersionedObject,
):
    fields = {
        'id': fields.StringField(),
        'name': fields.StringField(nullable=True),
        'group': fields.StringField(nullable=True),
        'tenant': fields.StringField(nullable=True),
        'config': heat_fields.JsonField(nullable=True),
        'created_at': fields.DateTimeField(read_only=True),
        'updated_at': fields.DateTimeField(nullable=True),
    }

    @staticmethod
    def _from_db_object(context, config, db_config):

        # SoftwareDeployment._from_db_object may attempt to load a None config
        if db_config is None:
            return None

        for field in config.fields:
            config[field] = db_config[field]
        config._context = context
        config.obj_reset_changes()
        return config

    @classmethod
    def create(cls, context, values):
        return cls._from_db_object(
            context, cls(), db_api.software_config_create(context, values))

    @classmethod
    def get_by_id(cls, context, config_id):
        return cls._from_db_object(
            context, cls(), db_api.software_config_get(context, config_id))

    @classmethod
    def get_all(cls, context, **kwargs):
        scs = db_api.software_config_get_all(context, **kwargs)
        return [cls._from_db_object(context, cls(), sc) for sc in scs]

    @classmethod
    def delete(cls, context, config_id):
        db_api.software_config_delete(context, config_id)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import exception
from heat.common import template_format
from heat.engine import properties
from heat.engine import resource
from heat.scaling import lbutils
from heat.tests import common
from heat.tests import generic_resource
from heat.tests import utils

lb_stack = '''
heat_template_version: 2013-05-23
resources:
  neutron_lb_1:
    type: Mock::Neutron::LB
  neutron_lb_2:
    type: Mock::Neutron::LB
  aws_lb_1:
    type: Mock::AWS::LB
  aws_lb_2:
    type: Mock::AWS::LB
  non_lb:
    type: Mock::Not::LB
'''


class MockedNeutronLB(generic_resource.GenericResource):
    properties_schema = {
        'members': properties.Schema(
            properties.Schema.LIST,
            update_allowed=True)
    }

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        return


class MockedAWSLB(generic_resource.GenericResource):
    properties_schema = {
        'Instances': properties.Schema(
            properties.Schema.LIST,
            update_allowed=True)
    }

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        return


class LBUtilsTest(common.HeatTestCase):
    # need to mock a group so that group_utils.get_member_refids will work
    # load_balancers is a dict of load_balancer objects, where each lb has
    # properties for checking

    def setUp(self):
        super(LBUtilsTest, self).setUp()
        resource._register_class('Mock::Neutron::LB', MockedNeutronLB)
        resource._register_class('Mock::AWS::LB', MockedAWSLB)
        resource._register_class('Mock::Not::LB',
                                 generic_resource.GenericResource)

        t = template_format.parse(lb_stack)
        self.stack = utils.parse_stack(t)

    def test_reload_aws_lb(self):
        id_list = ['ID1', 'ID2', 'ID3']

        lb1 = self.stack['aws_lb_1']
        lb2 = self.stack['aws_lb_2']
        lb1.action = mock.Mock(return_value=lb1.CREATE)
        lb2.action = mock.Mock(return_value=lb2.CREATE)
        lb1.handle_update = mock.Mock()
        lb2.handle_update = mock.Mock()
        prop_diff = {'Instances': id_list}

        lbutils.reconfigure_loadbalancers([lb1, lb2], id_list)

        # For verification's purpose, we just check the prop_diff
        lb1.handle_update.assert_called_with(mock.ANY, mock.ANY,
                                             prop_diff)
        lb2.handle_update.assert_called_with(mock.ANY, mock.ANY,
                                             prop_diff)

    def test_reload_neutron_lb(self):
        id_list = ['ID1', 'ID2', 'ID3']

        lb1 = self.stack['neutron_lb_1']
        lb2 = self.stack['neutron_lb_2']
        lb1.action = mock.Mock(return_value=lb1.CREATE)
        lb2.action = mock.Mock(return_value=lb2.CREATE)

        lb1.handle_update = mock.Mock()
        lb2.handle_update = mock.Mock()

        prop_diff = {'members': id_list}

        lbutils.reconfigure_loadbalancers([lb1, lb2], id_list)

        # For verification's purpose, we just check the prop_diff
        lb1.handle_update.assert_called_with(mock.ANY, mock.ANY,
                                             prop_diff)
        lb2.handle_update.assert_called_with(mock.ANY, mock.ANY,
                                             prop_diff)

    def test_reload_non_lb(self):
        id_list = ['ID1', 'ID2', 'ID3']
        non_lb = self.stack['non_lb']

        error = self.assertRaises(exception.Error,
                                  lbutils.reconfigure_loadbalancers,
                                  [non_lb], id_list)
        self.assertIn("Unsupported resource 'non_lb' in LoadBalancerNames",
                      str(error))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
# -*- coding: utf-8 -*-

import os
import re

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives.tables import Table


class TableFromText(Table):
    """Take input from a file and create a simple table.

    Example:

    .. table_from_text:: ../setup.cfg
    :header: Name,Plug-in
    :regex: (.*)=(.*)
    :start-after: heat.constraints =
    :end-before: heat.stack_lifecycle_plugins =
    :sort:

    file: input file relative to source directory
    header: comma separated list of column titles
    regex: regular expression to parse the source line into columns
    start-after: string to look for to start column data
    end-before: string to look for to end column data
    sort: flag for sorting column data
    """

    required_arguments = 1
    option_spec = {
        'header': directives.unchanged_required,
        'regex': directives.unchanged_required,
        'start-after': directives.unchanged_required,
        'end-before': directives.unchanged_required,
        'sort': directives.flag
    }

    def run(self):
        header = self.options.get('header').split(',')
        lines = self._get_lines()
        regex = self.options.get('regex')
        max_cols = len(header)

        table = nodes.table()
        tgroup = nodes.tgroup(max_cols)
        table += tgroup

        col_widths = self.get_column_widths(max_cols)
        if isinstance(col_widths, tuple):
            col_widths = col_widths[1]
        tgroup.extend(nodes.colspec(colwidth=col_width) for
                      col_width in col_widths)

        thead = nodes.thead()
        tgroup += thead
        thead += self.create_table_row(header)

        tbody = nodes.tbody()
        tgroup += tbody

        for row in lines:
            matched = re.search(regex, row)
            if matched:
                tbody += self.create_table_row(matched.groups())

        return [table]

    def create_table_row(self, row_cells):
        row = nodes.row()

        for cell in row_cells:
            entry = nodes.entry()
            row += entry
            entry += nodes.paragraph(text=cell.strip())

        return row

    def _get_lines(self):
        env = self.state.document.settings.env
        sourcefile = os.path.join(env.srcdir, self.arguments[0])
        startafter = self.options.get('start-after')
        endbefore = self.options.get('end-before')

        lines = [line.strip() for line in open(sourcefile)]

        if startafter is not None or endbefore is not None:
            includeline = not startafter
            result = []
            for line in lines:
                if not includeline and startafter and startafter in line:
                    includeline = True
                elif includeline and endbefore and endbefore in line:
                    includeline = False
                    break
                elif includeline:
                    result.append(line)
            lines = result

        if 'sort' in self.options:
            lines = sorted(lines)

        return lines


def setup(app):
    app.add_directive('table_from_text', TableFromText)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
APIs for dealing with input and output definitions for Software Configurations.
"""

import collections
import copy

from heat.common.i18n import _

from heat.common import exception
from heat.engine import constraints
from heat.engine import parameters
from heat.engine import properties


(
    IO_NAME, DESCRIPTION, TYPE,
    DEFAULT, REPLACE_ON_CHANGE, VALUE,
    ERROR_OUTPUT,
) = (
    'name', 'description', 'type',
    'default', 'replace_on_change', 'value',
    'error_output',
)

TYPES = (
    STRING_TYPE, NUMBER_TYPE, LIST_TYPE, JSON_TYPE, BOOLEAN_TYPE,
) = (
    'String', 'Number', 'CommaDelimitedList', 'Json', 'Boolean',
)


input_config_schema = {
    IO_NAME: properties.Schema(
        properties.Schema.STRING,
        _('Name of the input.'),
        required=True
    ),
    DESCRIPTION: properties.Schema(
        properties.Schema.STRING,
        _('Description of the input.')
    ),
    TYPE: properties.Schema(
        properties.Schema.STRING,
        _('Type of the value of the input.'),
        default=STRING_TYPE,
        constraints=[constraints.AllowedValues(TYPES)]
    ),
    DEFAULT: properties.Schema(
        properties.Schema.ANY,
        _('Default value for the input if none is specified.'),
    ),
    REPLACE_ON_CHANGE: properties.Schema(
        properties.Schema.BOOLEAN,
        _('Replace the deployment instead of updating it when the input '
          'value changes.'),
        default=False,
    ),
}

output_config_schema = {
    IO_NAME: properties.Schema(
        properties.Schema.STRING,
        _('Name of the output.'),
        required=True
    ),
    DESCRIPTION: properties.Schema(
        properties.Schema.STRING,
        _('Description of the output.')
    ),
    TYPE: properties.Schema(
        properties.Schema.STRING,
        _('Type of the value of the output.'),
        default=STRING_TYPE,
        constraints=[constraints.AllowedValues(TYPES)]
    ),
    ERROR_OUTPUT: properties.Schema(
        properties.Schema.BOOLEAN,
        _('Denotes that the deployment is in an error state if this '
          'output has a value.'),
        default=False
    )
}


class IOConfig(object):
    """Base class for the configuration data for a single input or output."""
    def __init__(self, **config):
        self._props = properties.Properties(self.schema, config)
        try:
            self._props.validate()
        except exception.StackValidationFailed as exc:
            raise ValueError(str(exc))

    def name(self):
        """Return the name of the input or output."""
        return self._props[IO_NAME]

    def as_dict(self):
        """Return a dict representation suitable for persisting."""
        return {k: v for k, v in self._props.items() if v is not None}

    def __repr__(self):
        return '%s(%s)' % (type(self).__name__,
                           ', '.join('%s=%s' % (k, repr(v))
                                     for k, v in self.as_dict().items()))


_no_value = object()


class InputConfig(IOConfig):
    """Class representing the configuration data for a single input."""
    schema = input_config_schema

    def __init__(self, value=_no_value, **config):
        if TYPE in config and DEFAULT in config:
            if config[DEFAULT] == '' and config[TYPE] != STRING_TYPE:
                # This is a legacy path, because default used to be of string
                # type, so we need to skip schema validation in this case.
                pass
            else:
                self.schema = copy.deepcopy(self.schema)
                config_param = parameters.Schema.from_dict(
                    'config', {'Type': config[TYPE]})
                self.schema[DEFAULT] = properties.Schema.from_parameter(
                    config_param)
        super(InputConfig, self).__init__(**config)
        self._value = value

    def default(self):
        """Return the default value of the input."""
        return self._props[DEFAULT]

    def replace_on_change(self):
        return self._props[REPLACE_ON_CHANGE]

    def as_dict(self):
        """Return a dict representation suitable for persisting."""
        d = super(InputConfig, self).as_dict()
        if not self._props[REPLACE_ON_CHANGE]:
            del d[REPLACE_ON_CHANGE]
        if self._value is not _no_value:
            d[VALUE] = self._value
        return d

    def input_data(self):
        """Return a name, value pair for the input."""
        value = self._value if self._value is not _no_value else None
        return self.name(), value


class OutputConfig(IOConfig):
    """Class representing the configuration data for a single output."""
    schema = output_config_schema

    def error_output(self):
        """Return True if the presence of the output indicates an error."""
        return self._props[ERROR_OUTPUT]


def check_io_schema_list(io_configs):
    """Check that an input or output schema list is of the correct type.

    Raises TypeError if the list itself is not a list, or if any of the
    members are not dicts.
    """
    if (
        not isinstance(io_configs, collections.abc.Sequence) or
        isinstance(io_configs, collections.abc.Mapping) or
        isinstance(io_configs, str)
    ):
        raise TypeError('Software Config I/O Schema must be in a list')

    if not all(
        isinstance(conf, collections.abc.Mapping) for conf in io_configs
    ):
        raise TypeError('Software Config I/O Schema must be a dict')



#
# Copyright (c) 2013 Docker, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import exception
from heat.common.i18n import _
from heat.common import template_format
from heat.engine import resource
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils

import testtools

from heat_docker.resources import docker_container
from heat_docker.tests import fake_docker_client as docker

docker_container.docker = docker


template = '''
{
    "AWSTemplateFormatVersion": "2010-09-09",
    "Description": "Test template",
    "Parameters": {},
    "Resources": {
        "Blog": {
            "Type": "DockerInc::Docker::Container",
            "Properties": {
                "image": "samalba/wordpress",
                "env": [
                    "FOO=bar"
                ]
            }
        }
    }
}
'''


class DockerContainerTest(common.HeatTestCase):

    def setUp(self):
        super(DockerContainerTest, self).setUp()
        for res_name, res_class in docker_container.resource_mapping().items():
            resource._register_class(res_name, res_class)

    def create_container(self, resource_name):
        t = template_format.parse(template)
        self.stack = utils.parse_stack(t)
        resource = docker_container.DockerContainer(
            resource_name,
            self.stack.t.resource_definitions(self.stack)[resource_name],
            self.stack)
        self.patchobject(resource, 'get_client',
                         return_value=docker.Client())
        self.assertIsNone(resource.validate())
        scheduler.TaskRunner(resource.create)()
        self.assertEqual((resource.CREATE, resource.COMPLETE),
                         resource.state)
        return resource

    def get_container_state(self, resource):
        client = resource.get_client()
        return client.inspect_container(resource.resource_id)['State']

    def test_resource_create(self):
        container = self.create_container('Blog')
        self.assertTrue(container.resource_id)
        running = self.get_container_state(container)['Running']
        self.assertIs(True, running)
        client = container.get_client()
        self.assertEqual(['samalba/wordpress'], client.pulled_images)
        self.assertIsNone(client.container_create[0]['name'])

    def test_create_with_name(self):
        t = template_format.parse(template)
        self.stack = utils.parse_stack(t)
        definition = self.stack.t.resource_definitions(self.stack)['Blog']
        props = t['Resources']['Blog']['Properties'].copy()
        props['name'] = 'super-blog'
        resource = docker_container.DockerContainer(
            'Blog', definition.freeze(properties=props), self.stack)
        self.patchobject(resource, 'get_client',
                         return_value=docker.Client())
        self.assertIsNone(resource.validate())
        scheduler.TaskRunner(resource.create)()
        self.assertEqual((resource.CREATE, resource.COMPLETE),
                         resource.state)
        client = resource.get_client()
        self.assertEqual(['samalba/wordpress'], client.pulled_images)
        self.assertEqual('super-blog', client.container_create[0]['name'])

    @mock.patch.object(docker_container.DockerContainer, 'get_client')
    def test_create_failed(self, test_client):
        t = template_format.parse(template)
        self.stack = utils.parse_stack(t)
        definition = self.stack.t.resource_definitions(self.stack)['Blog']
        props = t['Resources']['Blog']['Properties'].copy()

        mock_client = mock.Mock()
        mock_client.inspect_container.return_value = {
            "State": {
                "ExitCode": -1
            }
        }
        mock_client.logs.return_value = "Container startup failed"
        test_client.return_value = mock_client
        docker_res = docker_container.DockerContainer(
            'Blog', definition.freeze(properties=props), self.stack)
        exc = self.assertRaises(exception.ResourceInError,
                                docker_res.check_create_complete,
                                'foo')
        self.assertIn("Container startup failed", str(exc))

    def test_start_with_bindings_and_links(self):
        t = template_format.parse(template)
        self.stack = utils.parse_stack(t)
        definition = self.stack.t.resource_definitions(self.stack)['Blog']
        props = t['Resources']['Blog']['Properties'].copy()
        props['port_bindings'] = {'80/tcp': [{'HostPort': '80'}]}
        props['links'] = {'db': 'mysql'}
        resource = docker_container.DockerContainer(
            'Blog', definition.freeze(properties=props), self.stack)
        self.patchobject(resource, 'get_client',
                         return_value=docker.Client())
        self.assertIsNone(resource.validate())
        scheduler.TaskRunner(resource.create)()
        self.assertEqual((resource.CREATE, resource.COMPLETE),
                         resource.state)
        client = resource.get_client()
        self.assertEqual(['samalba/wordpress'], client.pulled_images)
        self.assertEqual({'db': 'mysql'}, client.container_start[0]['links'])
        self.assertEqual(
            {'80/tcp': [{'HostPort': '80'}]},
            client.container_start[0]['port_bindings'])

    def test_resource_attributes(self):
        container = self.create_container('Blog')
        # Test network info attributes
        self.assertEqual('172.17.42.1', container.FnGetAtt('network_gateway'))
        self.assertEqual('172.17.0.3', container.FnGetAtt('network_ip'))
        self.assertEqual('1080', container.FnGetAtt('network_tcp_ports'))
        self.assertEqual('', container.FnGetAtt('network_udp_ports'))
        # Test logs attributes
        self.assertEqual('---logs_begin---', container.FnGetAtt('logs_head'))
        self.assertEqual('---logs_end---', container.FnGetAtt('logs_tail'))
        # Test a non existing attribute
        self.assertRaises(exception.InvalidTemplateAttribute,
                          container.FnGetAtt, 'invalid_attribute')

    @testtools.skipIf(docker is None, 'docker-py not available')
    def test_resource_delete(self):
        container = self.create_container('Blog')
        scheduler.TaskRunner(container.delete)()
        self.assertEqual((container.DELETE, container.COMPLETE),
                         container.state)

        exists = True
        try:
            self.get_container_state(container)['Running']
        except docker.errors.APIError as error:
            if error.response.status_code == 404:
                exists = False
            else:
                raise

        self.assertIs(False, exists)

    @testtools.skipIf(docker is None, 'docker-py not available')
    def test_resource_delete_exception(self):
        response = mock.MagicMock()
        response.status_code = 404
        response.content = 'some content'

        container = self.create_container('Blog')
        self.patchobject(container.get_client(), 'kill',
                         side_effect=[docker.errors.APIError(
                             'Not found', response)])

        self.patchobject(container, '_get_container_status',
                         side_effect=[docker.errors.APIError(
                             'Not found', response)])
        scheduler.TaskRunner(container.delete)()
        container.get_client().kill.assert_called_once_with(
            container.resource_id)
        container._get_container_status.assert_called_once_with(
            container.resource_id)

    def test_resource_suspend_resume(self):
        container = self.create_container('Blog')
        # Test suspend
        scheduler.TaskRunner(container.suspend)()
        self.assertEqual((container.SUSPEND, container.COMPLETE),
                         container.state)
        running = self.get_container_state(container)['Running']
        self.assertIs(False, running)
        # Test resume
        scheduler.TaskRunner(container.resume)()
        self.assertEqual((container.RESUME, container.COMPLETE),
                         container.state)
        running = self.get_container_state(container)['Running']
        self.assertIs(True, running)

    def test_start_with_restart_policy_no(self):
        t = template_format.parse(template)
        self.stack = utils.parse_stack(t)
        definition = self.stack.t.resource_definitions(self.stack)['Blog']
        props = t['Resources']['Blog']['Properties'].copy()
        props['restart_policy'] = {'Name': 'no', 'MaximumRetryCount': 0}
        resource = docker_container.DockerContainer(
            'Blog', definition.freeze(properties=props), self.stack)
        get_client_mock = self.patchobject(resource, 'get_client')
        get_client_mock.return_value = docker.Client()
        self.assertIsNone(resource.validate())
        scheduler.TaskRunner(resource.create)()
        self.assertEqual((resource.CREATE, resource.COMPLETE),
                         resource.state)
        client = resource.get_client()
        self.assertEqual(['samalba/wordpress'], client.pulled_images)
        self.assertEqual({'Name': 'no', 'MaximumRetryCount': 0},
                         client.container_start[0]['restart_policy'])

    def test_start_with_restart_policy_on_failure(self):
        t = template_format.parse(template)
        self.stack = utils.parse_stack(t)
        definition = self.stack.t.resource_definitions(self.stack)['Blog']
        props = t['Resources']['Blog']['Properties'].copy()
        props['restart_policy'] = {'Name': 'on-failure',
                                   'MaximumRetryCount': 10}
        resource = docker_container.DockerContainer(
            'Blog', definition.freeze(properties=props), self.stack)
        get_client_mock = self.patchobject(resource, 'get_client')
        get_client_mock.return_value = docker.Client()
        self.assertIsNone(resource.validate())
        scheduler.TaskRunner(resource.create)()
        self.assertEqual((resource.CREATE, resource.COMPLETE),
                         resource.state)
        client = resource.get_client()
        self.assertEqual(['samalba/wordpress'], client.pulled_images)
        self.assertEqual({'Name': 'on-failure', 'MaximumRetryCount': 10},
                         client.container_start[0]['restart_policy'])

    def test_start_with_restart_policy_always(self):
        t = template_format.parse(template)
        self.stack = utils.parse_stack(t)
        definition = self.stack.t.resource_definitions(self.stack)['Blog']
        props = t['Resources']['Blog']['Properties'].copy()
        props['restart_policy'] = {'Name': 'always', 'MaximumRetryCount': 0}
        resource = docker_container.DockerContainer(
            'Blog', definition.freeze(properties=props), self.stack)
        get_client_mock = self.patchobject(resource, 'get_client')
        get_client_mock.return_value = docker.Client()
        self.assertIsNone(resource.validate())
        scheduler.TaskRunner(resource.create)()
        self.assertEqual((resource.CREATE, resource.COMPLETE),
                         resource.state)
        client = resource.get_client()
        self.assertEqual(['samalba/wordpress'], client.pulled_images)
        self.assertEqual({'Name': 'always', 'MaximumRetryCount': 0},
                         client.container_start[0]['restart_policy'])

    def test_start_with_caps(self):
        t = template_format.parse(template)
        self.stack = utils.parse_stack(t)
        definition = self.stack.t.resource_definitions(self.stack)['Blog']
        props = t['Resources']['Blog']['Properties'].copy()
        props['cap_add'] = ['NET_ADMIN']
        props['cap_drop'] = ['MKNOD']
        resource = docker_container.DockerContainer(
            'Blog', definition.freeze(properties=props), self.stack)
        get_client_mock = self.patchobject(resource, 'get_client')
        get_client_mock.return_value = docker.Client()
        self.assertIsNone(resource.validate())
        scheduler.TaskRunner(resource.create)()
        self.assertEqual((resource.CREATE, resource.COMPLETE),
                         resource.state)
        client = resource.get_client()
        self.assertEqual(['samalba/wordpress'], client.pulled_images)
        self.assertEqual(['NET_ADMIN'], client.container_start[0]['cap_add'])
        self.assertEqual(['MKNOD'], client.container_start[0]['cap_drop'])

    def test_start_with_read_only(self):
        t = template_format.parse(template)
        self.stack = utils.parse_stack(t)
        definition = self.stack.t.resource_definitions(self.stack)['Blog']
        props = t['Resources']['Blog']['Properties'].copy()
        props['read_only'] = True
        resource = docker_container.DockerContainer(
            'Blog', definition.freeze(properties=props), self.stack)
        get_client_mock = self.patchobject(resource, 'get_client')
        get_client_mock.return_value = docker.Client()
        get_client_mock.return_value.set_api_version('1.17')
        self.assertIsNone(resource.validate())
        scheduler.TaskRunner(resource.create)()
        self.assertEqual((resource.CREATE, resource.COMPLETE),
                         resource.state)
        client = resource.get_client()
        self.assertEqual(['samalba/wordpress'], client.pulled_images)
        self.assertIs(True, client.container_start[0]['read_only'])

    def arg_for_low_api_version(self, arg, value, low_version):
        t = template_format.parse(template)
        self.stack = utils.parse_stack(t)
        definition = self.stack.t.resource_definitions(self.stack)['Blog']
        props = t['Resources']['Blog']['Properties'].copy()
        props[arg] = value
        my_resource = docker_container.DockerContainer(
            'Blog', definition.freeze(properties=props), self.stack)
        get_client_mock = self.patchobject(my_resource, 'get_client')
        get_client_mock.return_value = docker.Client()
        get_client_mock.return_value.set_api_version(low_version)
        msg = self.assertRaises(docker_container.InvalidArgForVersion,
                                my_resource.validate)
        min_version = docker_container.MIN_API_VERSION_MAP[arg]
        args = dict(arg=arg, min_version=min_version)
        expected = _('"%(arg)s" is not supported for API version '
                     '< "%(min_version)s"') % args
        self.assertEqual(expected, str(msg))

    def test_start_with_read_only_for_low_api_version(self):
        self.arg_for_low_api_version('read_only', True, '1.16')

    def test_compare_version(self):
        self.assertEqual(docker_container.compare_version('1.17', '1.17'), 0)
        self.assertEqual(docker_container.compare_version('1.17', '1.16'), -1)
        self.assertEqual(docker_container.compare_version('1.17', '1.18'), 1)

    def test_create_with_cpu_shares(self):
        t = template_format.parse(template)
        self.stack = utils.parse_stack(t)
        definition = self.stack.t.resource_definitions(self.stack)['Blog']
        props = t['Resources']['Blog']['Properties'].copy()
        props['cpu_shares'] = 512
        my_resource = docker_container.DockerContainer(
            'Blog', definition.freeze(properties=props), self.stack)
        get_client_mock = self.patchobject(my_resource, 'get_client')
        get_client_mock.return_value = docker.Client()
        self.assertIsNone(my_resource.validate())
        scheduler.TaskRunner(my_resource.create)()
        self.assertEqual((my_resource.CREATE, my_resource.COMPLETE),
                         my_resource.state)
        client = my_resource.get_client()
        self.assertEqual(['samalba/wordpress'], client.pulled_images)
        self.assertEqual(512, client.container_create[0]['cpu_shares'])

    def test_create_with_cpu_shares_for_low_api_version(self):
        self.arg_for_low_api_version('cpu_shares', 512, '1.7')

    def test_start_with_mapping_devices(self):
        t = template_format.parse(template)
        self.stack = utils.parse_stack(t)
        definition = self.stack.t.resource_definitions(self.stack)['Blog']
        props = t['Resources']['Blog']['Properties'].copy()
        props['devices'] = (
            [{'path_on_host': '/dev/sda',
              'path_in_container': '/dev/xvdc',
              'permissions': 'r'},
             {'path_on_host': '/dev/mapper/a_bc-d',
              'path_in_container': '/dev/xvdd',
              'permissions': 'rw'}])
        my_resource = docker_container.DockerContainer(
            'Blog', definition.freeze(properties=props), self.stack)
        get_client_mock = self.patchobject(my_resource, 'get_client')
        get_client_mock.return_value = docker.Client()
        self.assertIsNone(my_resource.validate())
        scheduler.TaskRunner(my_resource.create)()
        self.assertEqual((my_resource.CREATE, my_resource.COMPLETE),
                         my_resource.state)
        client = my_resource.get_client()
        self.assertEqual(['samalba/wordpress'], client.pulled_images)
        self.assertEqual(['/dev/sda:/dev/xvdc:r',
                          '/dev/mapper/a_bc-d:/dev/xvdd:rw'],
                         client.container_start[0]['devices'])

    def test_start_with_mapping_devices_also_with_privileged(self):
        t = template_format.parse(template)
        self.stack = utils.parse_stack(t)
        definition = self.stack.t.resource_definitions(self.stack)['Blog']
        props = t['Resources']['Blog']['Properties'].copy()
        props['devices'] = (
            [{'path_on_host': '/dev/sdb',
              'path_in_container': '/dev/xvdc',
              'permissions': 'r'}])
        props['privileged'] = True
        my_resource = docker_container.DockerContainer(
            'Blog', definition.freeze(properties=props), self.stack)
        get_client_mock = self.patchobject(my_resource, 'get_client')
        get_client_mock.return_value = docker.Client()
        self.assertIsNone(my_resource.validate())
        scheduler.TaskRunner(my_resource.create)()
        self.assertEqual((my_resource.CREATE, my_resource.COMPLETE),
                         my_resource.state)
        client = my_resource.get_client()
        self.assertEqual(['samalba/wordpress'], client.pulled_images)
        self.assertNotIn('devices', client.container_start[0])

    def test_start_with_mapping_devices_for_low_api_version(self):
        value = ([{'path_on_host': '/dev/sda',
                   'path_in_container': '/dev/xvdc',
                   'permissions': 'rwm'}])
        self.arg_for_low_api_version('devices', value, '1.13')

    def test_start_with_mapping_devices_not_set_path_in_container(self):
        t = template_format.parse(template)
        self.stack = utils.parse_stack(t)
        definition = self.stack.t.resource_definitions(self.stack)['Blog']
        props = t['Resources']['Blog']['Properties'].copy()
        props['devices'] = [{'path_on_host': '/dev/sda',
                             'permissions': 'rwm'}]
        my_resource = docker_container.DockerContainer(
            'Blog', definition.freeze(properties=props), self.stack)
        get_client_mock = self.patchobject(my_resource, 'get_client')
        get_client_mock.return_value = docker.Client()
        self.assertIsNone(my_resource.validate())
        scheduler.TaskRunner(my_resource.create)()
        self.assertEqual((my_resource.CREATE, my_resource.COMPLETE),
                         my_resource.state)
        client = my_resource.get_client()
        self.assertEqual(['samalba/wordpress'], client.pulled_images)
        self.assertEqual(['/dev/sda:/dev/sda:rwm'],
                         client.container_start[0]['devices'])

    def test_create_with_cpu_set(self):
        t = template_format.parse(template)
        self.stack = utils.parse_stack(t)
        definition = self.stack.t.resource_definitions(self.stack)['Blog']
        props = t['Resources']['Blog']['Properties'].copy()
        props['cpu_set'] = '0-8,16-24,28'
        my_resource = docker_container.DockerContainer(
            'Blog', definition.freeze(properties=props), self.stack)
        get_client_mock = self.patchobject(my_resource, 'get_client')
        get_client_mock.return_value = docker.Client()
        self.assertIsNone(my_resource.validate())
        scheduler.TaskRunner(my_resource.create)()
        self.assertEqual((my_resource.CREATE, my_resource.COMPLETE),
                         my_resource.state)
        client = my_resource.get_client()
        self.assertEqual(['samalba/wordpress'], client.pulled_images)
        self.assertEqual('0-8,16-24,28',
                         client.container_create[0]['cpuset'])

    def test_create_with_cpu_set_for_low_api_version(self):
        self.arg_for_low_api_version('cpu_set', '0-8,^2', '1.11')



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from blazarclient import exception as client_exception
from oslo_utils.fixture import uuidsentinel as uuids

from heat.common import exception
from heat.common import template_format
from heat.engine.clients.os import blazar
from heat.engine.resources.openstack.blazar import host
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils


blazar_host_template = '''
heat_template_version: rocky

resources:
  test-host:
    type: OS::Blazar::Host
    properties:
      name: test-host
      extra_capability:
        gpu: true
'''

blazar_host_template_extra_capability = '''
heat_template_version: rocky

resources:
  test-host:
    type: OS::Blazar::Host
    properties:
      name: test-host
      extra_capability:
        gpu: true
        name: test-name
'''


class BlazarHostTestCase(common.HeatTestCase):

    def setUp(self):
        super(BlazarHostTestCase, self).setUp()

        self.host = {
            "id": uuids.id,
            "name": "test-host",
            "gpu": True,
            "hypervisor_hostname": "compute-1",
            "hypervisor_type": "QEMU",
            "hypervisor_version": 2010001,
            "cpu_info": "{"
                        "'arch': 'x86_64', 'model': 'cpu64-rhel6', "
                        "'vendor': 'Intel', 'features': "
                        "['pge', 'clflush', 'sep', 'syscall', 'msr', "
                        "'vmx', 'cmov', 'nx', 'pat', 'lm', 'tsc', "
                        "'fpu', 'fxsr', 'pae', 'mmx', 'cx8', 'mce', "
                        "'de', 'mca', 'pse', 'pni', 'apic', 'sse', "
                        "'lahf_lm', 'sse2', 'hypervisor', 'cx16', "
                        "'pse36', 'mttr', 'x2apic'], "
                        "'topology': {'cores': 1, 'cells': 1, 'threads': 1, "
                        "'sockets': 4}}",
            "memory_mb": 8192,
            "local_gb": 100,
            "vcpus": 2,
            "service_name": "compute-1",
            "reservable": True,
            "trust_id": uuids.trust_id,
            "created_at": "2020-01-01 08:00",
            "updated_at": "2020-01-01 12:00",
            "extra_capability": "foo"
        }
        t = template_format.parse(blazar_host_template)
        self.stack = utils.parse_stack(t)
        resource_defns = self.stack.t.resource_definitions(self.stack)
        self.rsrc_defn = resource_defns['test-host']
        self.client = mock.Mock()
        self.patchobject(blazar.BlazarClientPlugin, 'client',
                         return_value=self.client)

    def _create_resource(self, name, snippet, stack):
        self.client.host.create.return_value = self.host
        return host.Host(name, snippet, stack)

    def test_host_create(self):
        host_resource = self._create_resource('host', self.rsrc_defn,
                                              self.stack)

        self.assertEqual(self.host['name'],
                         host_resource.properties.get(host.Host.NAME))

        scheduler.TaskRunner(host_resource.create)()
        self.assertEqual(uuids.id, host_resource.resource_id)
        self.assertEqual((host_resource.CREATE, host_resource.COMPLETE),
                         host_resource.state)
        self.assertEqual('host', host_resource.entity)
        self.client.host.create.assert_called_once_with(
            name=self.host['name'], gpu=self.host['gpu'])

    def test_host_delete(self):
        host_resource = self._create_resource('host', self.rsrc_defn,
                                              self.stack)

        scheduler.TaskRunner(host_resource.create)()
        self.client.host.delete.return_value = None
        self.client.host.get.side_effect = [
            'host_obj', client_exception.BlazarClientException(code=404)]
        scheduler.TaskRunner(host_resource.delete)()
        self.assertEqual((host_resource.DELETE, host_resource.COMPLETE),
                         host_resource.state)
        self.client.host.delete.assert_called_once_with(uuids.id)

    def test_host_delete_not_found(self):
        host_resource = self._create_resource('host', self.rsrc_defn,
                                              self.stack)

        scheduler.TaskRunner(host_resource.create)()
        self.client.host.delete.side_effect = client_exception.\
            BlazarClientException(code=404)
        self.client.host.get.side_effect = client_exception.\
            BlazarClientException(code=404)
        scheduler.TaskRunner(host_resource.delete)()
        self.assertEqual((host_resource.DELETE, host_resource.COMPLETE),
                         host_resource.state)

    def test_parse_extra_capability(self):
        t = template_format.parse(blazar_host_template_extra_capability)
        stack = utils.parse_stack(t)
        resource_defns = self.stack.t.resource_definitions(stack)
        rsrc_defn = resource_defns['test-host']
        host_resource = self._create_resource('host', rsrc_defn, stack)
        args = dict((k, v) for k, v in host_resource.properties.items()
                    if v is not None)
        parsed_args = host_resource._parse_extra_capability(args)
        self.assertEqual({'gpu': True, 'name': 'test-host'}, parsed_args)

    def test_resolve_attributes(self):
        host_resource = self._create_resource('host', self.rsrc_defn,
                                              self.stack)

        scheduler.TaskRunner(host_resource.create)()
        self.client.host.get.return_value = self.host
        self.assertEqual(self.host['vcpus'],
                         host_resource._resolve_attribute(host.Host.VCPUS))

    def test_resolve_attributes_not_found(self):
        host_resource = self._create_resource('host', self.rsrc_defn,
                                              self.stack)

        scheduler.TaskRunner(host_resource.create)()
        self.client.host.get.return_value = self.host
        self.assertRaises(exception.InvalidTemplateAttribute,
                          host_resource._resolve_attribute,
                          'invalid_attribute')
